@@ -1,6 +1,21 @@
+/*
+   Copyright (c) 2016,2017 MariaDB
+
+   This program is free software; you can redistribute it and/or modify
+   it under the terms of the GNU General Public License as published by
+   the Free Software Foundation; version 2 of the License.
+
+   This program is distributed in the hope that it will be useful,
+   but WITHOUT ANY WARRANTY; without even the implied warranty of
+   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+   GNU General Public License for more details.
+
+   You should have received a copy of the GNU General Public License
+   along with this program; if not, write to the Free Software
+   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA */
+
+#include "mariadb.h"
 #include "item_windowfunc.h" 
-#include "my_dbug.h"
-#include "my_global.h"
 #include "sql_select.h" // test if group changed
 
 
@@ -13,7 +28,7 @@ Item_window_func::resolve_window_name(THD *thd)
     return false;
   }
   DBUG_ASSERT(window_name != NULL && window_spec == NULL);
-  char *ref_name= window_name->str;
+  const char *ref_name= window_name->str;
 
   /* !TODO: Add the code to resolve ref_name in outer queries */ 
   /* 
@@ -26,7 +41,7 @@ Item_window_func::resolve_window_name(THD *thd)
   Window_spec *win_spec;
   while((win_spec= it++))
   {
-    char *win_spec_name= win_spec->name();
+    const char *win_spec_name= win_spec->name();
     if (win_spec_name &&
         my_strcasecmp(system_charset_info, ref_name, win_spec_name) == 0)
     {
@@ -93,6 +108,7 @@ Item_window_func::fix_fields(THD *thd, Item **ref)
     my_error(ER_NO_ORDER_LIST_IN_WINDOW_SPEC, MYF(0), window_func()->func_name());
     return true;
   }
+
   /*
     TODO: why the last parameter is 'ref' in this call? What if window_func
     decides to substitute itself for something else and does *ref=.... ? 
@@ -104,7 +120,6 @@ Item_window_func::fix_fields(THD *thd, Item **ref)
 
   const_item_cache= false;
   with_window_func= true;
-  with_sum_func= false;
 
   if (fix_length_and_dec())
     return TRUE;
@@ -154,10 +169,46 @@ void Item_window_func::split_sum_func(THD *thd, Ref_ptr_array ref_pointer_array,
   window_func()->setup_caches(thd);
 }
 
+bool Item_window_func::check_result_type_of_order_item()
+{
+  switch (window_func()->sum_func()) {
+  case Item_sum::PERCENTILE_CONT_FUNC:
+  {
+    Item_result rtype= window_spec->order_list->first->item[0]->cmp_type();
+    // TODO (varun) : support date type in percentile_cont function
+    if (rtype != REAL_RESULT && rtype != INT_RESULT &&
+        rtype != DECIMAL_RESULT && rtype != TIME_RESULT)
+    {
+      my_error(ER_WRONG_TYPE_FOR_PERCENTILE_FUNC, MYF(0), window_func()->func_name());
+      return true;
+    }
+    return false;
+  }
+  case Item_sum::PERCENTILE_DISC_FUNC:
+  {
+    Item *src_item= window_spec->order_list->first->item[0];
+    Item_result rtype= src_item->cmp_type();
+    // TODO-10.5: Fix MDEV-20280 PERCENTILE_DISC() rejects temporal and string input
+    if (rtype != REAL_RESULT && rtype != INT_RESULT && rtype != DECIMAL_RESULT)
+    {
+      my_error(ER_WRONG_TYPE_FOR_PERCENTILE_FUNC, MYF(0), window_func()->func_name());
+      return true;
+    }
+    Item_sum_percentile_disc *func=
+      static_cast<Item_sum_percentile_disc*>(window_func());
+    func->set_handler(src_item->type_handler());
+    func->Type_std_attributes::set(src_item);
+    Type_std_attributes::set(src_item);
+    return false;
+  }
+  default:
+    break;
+  }
+  return FALSE;
+}
 
 /*
   This must be called before attempting to compute the window function values.
-
   @detail
     If we attempt to do it in fix_fields(), partition_fields will refer
     to the original window function arguments.
@@ -180,6 +231,71 @@ void Item_sum_dense_rank::setup_window_func(THD *thd, Window_spec *window_spec)
   clear();
 }
 
+void Item_sum_percentile_disc::setup_window_func(THD *thd, Window_spec *window_spec)
+{
+  order_item= window_spec->order_list->first->item[0];
+  if (!(value= order_item->get_cache(thd)))
+    return;
+  value->setup(thd, order_item);
+  value->store(order_item);
+}
+
+void Item_sum_percentile_cont::setup_window_func(THD *thd, Window_spec *window_spec)
+{
+  order_item= window_spec->order_list->first->item[0];
+  /* TODO(varun): need to discuss and finalise what type should we
+     return for percentile cont functions
+  */
+  if (!(ceil_value= order_item->get_cache(thd)))
+    return;
+  ceil_value->setup(thd, order_item);
+  ceil_value->store(order_item);
+
+  if (!(floor_value= order_item->get_cache(thd)))
+    return;
+  floor_value->setup(thd, order_item);
+  floor_value->store(order_item);
+}
+bool Item_sum_percentile_cont::fix_fields(THD *thd, Item **ref)
+{
+  bool res;
+  res= Item_sum_num::fix_fields(thd, ref);
+  if (res)
+    return res;
+
+  switch(args[0]->cmp_type())
+  {
+    case DECIMAL_RESULT:
+    case REAL_RESULT:
+    case INT_RESULT:
+      break;
+    default:
+      my_error(ER_WRONG_TYPE_OF_ARGUMENT, MYF(0), func_name());
+      return TRUE;
+  }
+  return res;
+}
+bool Item_sum_percentile_disc::fix_fields(THD *thd, Item **ref)
+{
+  bool res;
+  res= Item_sum_num::fix_fields(thd, ref);
+  if (res)
+    return res;
+
+  switch(args[0]->cmp_type())
+  {
+    case DECIMAL_RESULT:
+    case REAL_RESULT:
+    case INT_RESULT:
+      break;
+    default:
+      my_error(ER_WRONG_TYPE_OF_ARGUMENT, MYF(0), func_name());
+      return TRUE;
+  }
+  return res;
+
+}
+
 bool Item_sum_dense_rank::add()
 {
   if (peer_tracker->check_if_next_group() || first_add)
@@ -232,46 +348,19 @@ bool Item_sum_hybrid_simple::fix_fields(THD *thd, Item **ref)
 
   for (uint i= 0; i < arg_count; i++)
   {
-    Item *item= args[i];
     // 'item' can be changed during fix_fields
-    if ((!item->fixed && item->fix_fields(thd, args + i)) ||
-        (item= args[i])->check_cols(1))
+    if (args[i]->fix_fields_if_needed_for_scalar(thd, &args[i]))
       return TRUE;
-    with_window_func|= item->with_window_func;
   }
-  Type_std_attributes::set(args[0]);
-  for (uint i= 0; i < arg_count && !with_subselect; i++)
-    with_subselect= with_subselect || args[i]->with_subselect;
 
-  Item *item2= args[0]->real_item();
-  if (item2->type() == Item::FIELD_ITEM)
-    set_handler_by_field_type(((Item_field*) item2)->field->type());
-  else if (args[0]->cmp_type() == TIME_RESULT)
-    set_handler_by_field_type(item2->field_type());
-  else
-    set_handler_by_result_type(item2->result_type(),
-                               max_length, collation.collation);
+  for (uint i= 0; i < arg_count && !m_with_subquery; i++)
+    m_with_subquery|= args[i]->with_subquery();
 
-  switch (Item_sum_hybrid_simple::result_type()) {
-  case INT_RESULT:
-  case DECIMAL_RESULT:
-  case STRING_RESULT:
-    break;
-  case REAL_RESULT:
-    max_length= float_length(decimals);
-    break;
-  case ROW_RESULT:
-  case TIME_RESULT:
-    DBUG_ASSERT(0); // XXX(cvicentiu) Should this never happen?
-    return TRUE;
-  };
-  setup_hybrid(thd, args[0]);
-  /* MIN/MAX can return NULL for empty set indepedent of the used column */
-  maybe_null= 1;
-  result_field=0;
-  null_value=1;
   if (fix_length_and_dec())
-    return TRUE;
+    return true;
+
+  setup_hybrid(thd, args[0]);
+  result_field=0;
 
   if (check_sum_func(thd, ref))
     return TRUE;
@@ -283,6 +372,14 @@ bool Item_sum_hybrid_simple::fix_fields(THD *thd, Item **ref)
   return FALSE;
 }
 
+
+bool Item_sum_hybrid_simple::fix_length_and_dec()
+{
+  maybe_null= null_value= true;
+  return args[0]->type_handler()->Item_sum_hybrid_fix_length_and_dec(this);
+}
+
+
 bool Item_sum_hybrid_simple::add()
 {
   value->store(args[0]);
@@ -293,7 +390,7 @@ bool Item_sum_hybrid_simple::add()
 
 void Item_sum_hybrid_simple::setup_hybrid(THD *thd, Item *item)
 {
-  if (!(value= Item_cache::get_cache(thd, item, item->cmp_type())))
+  if (!(value= item->get_cache(thd)))
     return;
   value->setup(thd, item);
   value->store(item);
@@ -347,6 +444,25 @@ Item_sum_hybrid_simple::val_str(String *str)
   return retval;
 }
 
+bool Item_sum_hybrid_simple::val_native(THD *thd, Native *to)
+{
+  DBUG_ASSERT(fixed == 1);
+  if (null_value)
+    return true;
+  return val_native_from_item(thd, value, to);
+}
+
+bool Item_sum_hybrid_simple::get_date(THD *thd, MYSQL_TIME *ltime, date_mode_t fuzzydate)
+{
+  DBUG_ASSERT(fixed == 1);
+  if (null_value)
+    return true;
+  bool retval= value->get_date(thd, ltime, fuzzydate);
+  if ((null_value= value->null_value))
+    DBUG_ASSERT(retval == true);
+  return retval;
+}
+
 Field *Item_sum_hybrid_simple::create_tmp_field(bool group, TABLE *table)
 {
   DBUG_ASSERT(0);
@@ -355,7 +471,7 @@ Field *Item_sum_hybrid_simple::create_tmp_field(bool group, TABLE *table)
 
 void Item_sum_hybrid_simple::reset_field()
 {
-  switch(Item_sum_hybrid_simple::result_type()) {
+  switch(result_type()) {
   case STRING_RESULT:
   {
     char buff[MAX_FIELD_WIDTH];
@@ -410,11 +526,11 @@ void Item_sum_hybrid_simple::reset_field()
   }
   case DECIMAL_RESULT:
   {
-    my_decimal value_buff, *arg_dec= args[0]->val_decimal(&value_buff);
+    VDec arg_dec(args[0]);
 
     if (maybe_null)
     {
-      if (args[0]->null_value)
+      if (arg_dec.is_null())
         result_field->set_null();
       else
         result_field->set_notnull();
@@ -423,9 +539,7 @@ void Item_sum_hybrid_simple::reset_field()
       We must store zero in the field as we will use the field value in
       add()
     */
-    if (!arg_dec)                               // Null
-      arg_dec= &decimal_zero;
-    result_field->store_decimal(arg_dec);
+    result_field->store_decimal(arg_dec.ptr_or(&decimal_zero));
     break;
   }
   case ROW_RESULT:
@@ -441,10 +555,29 @@ void Item_sum_hybrid_simple::update_field()
 
 void Item_window_func::print(String *str, enum_query_type query_type)
 {
+  if (only_single_element_order_list())
+  {
+    print_for_percentile_functions(str, query_type);
+    return;
+  }
   window_func()->print(str, query_type);
   str->append(" over ");
-  if (!window_spec)
+#ifndef DBUG_OFF
+  if (!window_spec) // one can call dbug_print_item() anytime in gdb
     str->append(window_name);
   else
-    window_spec->print(str, query_type);
+#endif
+  window_spec->print(str, query_type);
+}
+void Item_window_func::print_for_percentile_functions(String *str, enum_query_type query_type)
+{
+  window_func()->print(str, query_type);
+  str->append(" within group ");
+  str->append('(');
+  window_spec->print_order(str,query_type);
+  str->append(')');
+  str->append(" over ");
+  str->append('(');
+  window_spec->print_partition(str,query_type);
+  str->append(')');
 }
