@@ -18,9 +18,8 @@
   on Windows.
 */
 #define DONT_DEFINE_VOID
-#include <my_global.h>
+#include "mariadb.h"
 #include <my_getopt.h>
-#include <my_sys.h>
 #include <m_string.h>
 
 #include <windows.h>
@@ -40,7 +39,8 @@ struct IUnknown;
 
 extern "C" const char* mysql_bootstrap_sql[];
 
-char default_os_user[]= "NT AUTHORITY\\NetworkService";
+static char default_os_user[]= "NT AUTHORITY\\NetworkService";
+static char default_datadir[MAX_PATH];
 static int create_db_instance();
 static uint opt_silent;
 static char datadir_buffer[FN_REFLEN];
@@ -96,9 +96,7 @@ static struct my_option my_long_options[]=
 
 
 static my_bool
-get_one_option(int optid, 
-   const struct my_option *opt __attribute__ ((unused)),
-   char *argument __attribute__ ((unused)))
+get_one_option(int optid, const struct my_option *,  char *)
 {
   DBUG_ENTER("get_one_option");
   switch (optid) {
@@ -112,7 +110,7 @@ get_one_option(int optid,
 }
 
 
-static void die(const char *fmt, ...)
+ATTRIBUTE_NORETURN  static void die(const char *fmt, ...)
 {
   va_list args;
   DBUG_ENTER("die");
@@ -172,8 +170,27 @@ int main(int argc, char **argv)
     exit(error);
   if (!opt_datadir)
   {
-    my_print_help(my_long_options);
-    die("parameter --datadir=# is mandatory");
+    /*
+      Figure out default data directory. It "data" directory, next to "bin" directory, where
+      mysql_install_db.exe resides.
+    */
+    strcpy(default_datadir, self_name);
+    p = strrchr(default_datadir, FN_LIBCHAR);
+    if (p)
+    {
+      *p= 0;
+      p= strrchr(default_datadir, FN_LIBCHAR);
+      if (p)
+        *p= 0;
+    }
+    if (!p)
+    {
+      die("--datadir option not provided, and default datadir not found");
+      my_print_help(my_long_options);
+    }
+    strcat_s(default_datadir, "\\data");
+    opt_datadir= default_datadir;
+    printf("Default data directory is %s\n",opt_datadir);
   }
 
   /* Print some help on errors */
@@ -201,7 +218,7 @@ int main(int argc, char **argv)
     die("database creation failed");
   }
 
-  printf("Creation of the database was successful");
+  printf("Creation of the database was successful\n");
   return 0;
 }
 
@@ -293,7 +310,7 @@ static int create_myini()
   FILE *myini= fopen("my.ini","wt");
   if (!myini)
   {
-    die("Cannot create my.ini in data directory");
+    die("Can't create my.ini in data directory");
   }
 
   /* Write out server settings. */
@@ -311,7 +328,7 @@ static int create_myini()
 
   if (enable_named_pipe)
   {
-    fprintf(myini,"enable-named-pipe\n");
+    fprintf(myini,"named-pipe=ON\n");
   }
 
   if (opt_socket && opt_socket[0])
@@ -346,17 +363,20 @@ static int create_myini()
 
 
 static const char update_root_passwd_part1[]=
-  "UPDATE mysql.user SET Password = PASSWORD(";
+  "UPDATE mysql.global_priv SET priv=json_set(priv,"
+  "'$.password_last_changed', UNIX_TIMESTAMP(),"
+  "'$.plugin','mysql_native_password',"
+  "'$.authentication_string',PASSWORD(";
 static const char update_root_passwd_part2[]=
-  ") where User='root';\n";
-static const char remove_default_user_cmd[]= 
+  ")) where User='root';\n";
+static const char remove_default_user_cmd[]=
   "DELETE FROM mysql.user where User='';\n";
 static const char allow_remote_root_access_cmd[]=
-  "CREATE TEMPORARY TABLE tmp_user LIKE user;\n"
-  "INSERT INTO tmp_user SELECT * from user where user='root' "
+  "CREATE TEMPORARY TABLE tmp_user LIKE global_priv;\n"
+  "INSERT INTO tmp_user SELECT * from global_priv where user='root' "
     " AND host='localhost';\n"
   "UPDATE tmp_user SET host='%';\n"
-  "INSERT INTO user SELECT * FROM tmp_user;\n"
+  "INSERT INTO global_priv SELECT * FROM tmp_user;\n"
   "DROP TABLE tmp_user;\n";
 static const char end_of_script[]="-- end.";
 
@@ -406,8 +426,8 @@ static int register_service()
 
 static void clean_directory(const char *dir)
 {
-  char dir2[MAX_PATH + 4]= {};
-  snprintf(dir2, MAX_PATH+2, "%s\\*", dir);
+  char dir2[MAX_PATH+2];
+  *(strmake_buf(dir2, dir)+1)= 0;
 
   SHFILEOPSTRUCT fileop;
   fileop.hwnd= NULL;    /* no status display */
@@ -445,7 +465,6 @@ static int set_directory_permissions(const char *dir, const char *os_user)
   ACL* pOldDACL;
   SECURITY_DESCRIPTOR* pSD= NULL; 
   EXPLICIT_ACCESS ea={0};
-  BOOL isWellKnownSID= FALSE;
   WELL_KNOWN_SID_TYPE wellKnownSidType = WinNullSid;
   PSID pSid= NULL;
 
@@ -512,7 +531,7 @@ static int set_directory_permissions(const char *dir, const char *os_user)
   ea.grfInheritance= CONTAINER_INHERIT_ACE|OBJECT_INHERIT_ACE; 
   ea.Trustee.TrusteeType= TRUSTEE_IS_UNKNOWN; 
   ACL* pNewDACL= 0; 
-  DWORD err= SetEntriesInAcl(1,&ea,pOldDACL,&pNewDACL); 
+  SetEntriesInAcl(1,&ea,pOldDACL,&pNewDACL);
   if (pNewDACL)
   {
     SetSecurityInfo(hDir,SE_FILE_OBJECT,DACL_SECURITY_INFORMATION,NULL, NULL,
@@ -527,27 +546,6 @@ static int set_directory_permissions(const char *dir, const char *os_user)
 }
 
 
-/* 
-  Give directory permissions for special service user NT SERVICE\servicename
-  this user is available only on Win7 and later.
-*/
-
-void grant_directory_permissions_to_service()
-{
-  char service_user[MAX_PATH+ 12];
-  OSVERSIONINFO info;
-  info.dwOSVersionInfoSize= sizeof(info);
-  GetVersionEx(&info);
-  if (info.dwMajorVersion >6 || 
-    (info.dwMajorVersion== 6 && info.dwMinorVersion > 0)
-    && opt_service)
-  {
-    my_snprintf(service_user,sizeof(service_user), "NT SERVICE\\%s", 
-      opt_service);
-    set_directory_permissions(opt_datadir, service_user);
-  }
-}
-
 
 /* Create database instance (including registering as service etc) .*/
 
@@ -558,7 +556,7 @@ static int create_db_instance()
   DWORD cwd_len= MAX_PATH;
   char cmdline[3*MAX_PATH];
   FILE *in;
-  bool created_datadir= false;
+  bool cleanup_datadir= true;
   DWORD last_error;
 
   verbose("Running bootstrap");
@@ -567,11 +565,7 @@ static int create_db_instance()
 
   /* Create datadir and datadir/mysql, if they do not already exist. */
 
-  if (CreateDirectory(opt_datadir, NULL))
-  {
-    created_datadir= true;
-  }
-  else if (GetLastError() != ERROR_ALREADY_EXISTS)
+  if (!CreateDirectory(opt_datadir, NULL) && (GetLastError() != ERROR_ALREADY_EXISTS))
   {
     last_error = GetLastError();
     switch(last_error)
@@ -608,11 +602,9 @@ static int create_db_instance()
     }
   }
 
-  if (!PathIsDirectoryEmpty(opt_datadir))
+  if (PathIsDirectoryEmpty(opt_datadir))
   {
-    fprintf(stderr,"ERROR : Data directory %s is not empty."
-        " Only new or empty existing directories are accepted for --datadir\n",opt_datadir);
-    exit(1);
+    cleanup_datadir= false;
   }
 
   if (!CreateDirectory("mysql",NULL))
@@ -656,11 +648,11 @@ static int create_db_instance()
 
   if (setvbuf(in, NULL, _IONBF, 0))
   {
-    verbose("WARNING: Cannot disable buffering on mysqld's stdin");
+    verbose("WARNING: Can't disable buffering on mysqld's stdin");
   }
   if (fwrite("use mysql;\n",11,1, in) != 1)
   {
-    verbose("ERROR: Cannot write to mysqld's stdin");
+    verbose("ERROR: Can't write to mysqld's stdin");
     ret= 1;
     goto end;
   }
@@ -671,7 +663,7 @@ static int create_db_instance()
     /* Write the bootstrap script to stdin. */
     if (fwrite(mysql_bootstrap_sql[i], strlen(mysql_bootstrap_sql[i]), 1, in) != 1)
     {
-      verbose("ERROR: Cannot write to mysqld's stdin");
+      verbose("ERROR: Can't write to mysqld's stdin");
       ret= 1;
       goto end;
     }
@@ -735,18 +727,15 @@ static int create_db_instance()
   if (opt_service && opt_service[0])
   {
     ret= register_service();
-    grant_directory_permissions_to_service();
     if (ret)
       goto end;
   }
 
 end:
-  if (ret)
+  if (ret && cleanup_datadir)
   {
     SetCurrentDirectory(cwd);
     clean_directory(opt_datadir);
-    if (created_datadir)
-      RemoveDirectory(opt_datadir);
   }
   return ret;
 }
