@@ -13,7 +13,7 @@
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1335  USA */
 
-#include <my_global.h>
+#include "mariadb.h"
 #include "sql_priv.h"
 #include "unireg.h"
 #include "sql_help.h"
@@ -87,15 +87,19 @@ enum enum_used_fields
 static bool init_fields(THD *thd, TABLE_LIST *tables,
 			struct st_find_field *find_fields, uint count)
 {
-  Name_resolution_context *context= &thd->lex->select_lex.context;
+  Name_resolution_context *context= &thd->lex->first_select_lex()->context;
   DBUG_ENTER("init_fields");
   context->resolve_in_table_list_only(tables);
   for (; count-- ; find_fields++)
   {
+    LEX_CSTRING field_name= {find_fields->field_name,
+                             strlen(find_fields->field_name) };
     /* We have to use 'new' here as field will be re_linked on free */
-    Item_field *field= new (thd->mem_root) Item_field(thd, context,
-                                      "mysql", find_fields->table_name,
-                                      find_fields->field_name);
+    Item_field *field= (new (thd->mem_root)
+                        Item_field(thd, context,
+                                   "mysql",
+                                   find_fields->table_name,
+                                   &field_name));
     if (!(find_fields->field= find_field_in_tables(thd, field, tables, NULL,
 						   0, REPORT_ALL_ERRORS, 1,
                                                    TRUE)))
@@ -198,7 +202,7 @@ int search_topics(THD *thd, TABLE *topics, struct st_find_field *find_fields,
                        FALSE))
     DBUG_RETURN(0);
 
-  while (!read_record_info.read_record(&read_record_info))
+  while (!read_record_info.read_record())
   {
     if (!select->cond->val_int())		// Doesn't match like
       continue;
@@ -242,7 +246,7 @@ int search_keyword(THD *thd, TABLE *keywords,
                        FALSE))
     DBUG_RETURN(0);
 
-  while (!read_record_info.read_record(&read_record_info) && count<2)
+  while (!read_record_info.read_record() && count<2)
   {
     if (!select->cond->val_int())		// Dosn't match like
       continue;
@@ -376,7 +380,7 @@ int search_categories(THD *thd, TABLE *categories,
   if (init_read_record(&read_record_info, thd, categories, select, NULL,
                        1, 0, FALSE))
     DBUG_RETURN(0);
-  while (!read_record_info.read_record(&read_record_info))
+  while (!read_record_info.read_record())
   {
     if (select && !select->cond->val_int())
       continue;
@@ -414,7 +418,7 @@ void get_all_items_for_category(THD *thd, TABLE *items, Field *pfname,
                        FALSE))
     DBUG_VOID_RETURN;
 
-  while (!read_record_info.read_record(&read_record_info))
+  while (!read_record_info.read_record())
   {
     if (!select->cond->val_int())
       continue;
@@ -607,15 +611,15 @@ int send_variant_2_list(MEM_ROOT *mem_root, Protocol *protocol,
 SQL_SELECT *prepare_simple_select(THD *thd, Item *cond,
 				  TABLE *table, int *error)
 {
-  if (!cond->fixed)
-    cond->fix_fields(thd, &cond);	// can never fail
+  cond->fix_fields_if_needed(thd, &cond);  // can never fail
 
   /* Assume that no indexes cover all required fields */
   table->covering_keys.clear_all();
 
   SQL_SELECT *res= make_select(table, 0, 0, cond, 0, 0, error);
-  if (*error || (res && res->check_quick(thd, 0, HA_POS_ERROR)) ||
-      (res && res->quick && res->quick->reset()))
+  if (unlikely(*error) ||
+      (likely(res) && unlikely(res->check_quick(thd, 0, HA_POS_ERROR))) ||
+      (likely(res) && res->quick && unlikely(res->quick->reset())))
   {
     delete res;
     res=0;
@@ -641,7 +645,7 @@ SQL_SELECT *prepare_simple_select(THD *thd, Item *cond,
     #  created SQL_SELECT
 */
 
-SQL_SELECT *prepare_select_for_name(THD *thd, const char *mask, uint mlen,
+SQL_SELECT *prepare_select_for_name(THD *thd, const char *mask, size_t mlen,
 				    TABLE_LIST *tables, TABLE *table,
 				    Field *pfname, int *error)
 {
@@ -650,11 +654,11 @@ SQL_SELECT *prepare_select_for_name(THD *thd, const char *mask, uint mlen,
     Item_func_like(thd,
                    new (mem_root)
                    Item_field(thd, pfname),
-                   new (mem_root) Item_string(thd, mask, mlen,
+                   new (mem_root) Item_string(thd, mask, (uint)mlen,
                                               pfname->charset()),
                    new (mem_root) Item_string_ascii(thd, "\\"),
                    FALSE);
-  if (thd->is_fatal_error)
+  if (unlikely(thd->is_fatal_error))
     return 0;					// OOM
   return prepare_simple_select(thd, cond, table, error);
 }
@@ -682,23 +686,19 @@ static bool mysqld_help_internal(THD *thd, const char *mask)
   List<String> topics_list, categories_list, subcategories_list;
   String name, description, example;
   int count_topics, count_categories, error;
-  uint mlen= strlen(mask);
+  size_t mlen= strlen(mask);
   size_t i;
   MEM_ROOT *mem_root= thd->mem_root;
+  LEX_CSTRING MYSQL_HELP_TOPIC_NAME=    {STRING_WITH_LEN("help_topic") };
+  LEX_CSTRING MYSQL_HELP_CATEGORY_NAME= {STRING_WITH_LEN("help_category") };
+  LEX_CSTRING MYSQL_HELP_RELATION_NAME= {STRING_WITH_LEN("help_relation") };
+  LEX_CSTRING MYSQL_HELP_KEYWORD_NAME=  {STRING_WITH_LEN("help_keyword") };
   DBUG_ENTER("mysqld_help");
 
-  tables[0].init_one_table(C_STRING_WITH_LEN("mysql"),
-                           C_STRING_WITH_LEN("help_topic"),
-                           "help_topic", TL_READ);
-  tables[1].init_one_table(C_STRING_WITH_LEN("mysql"),
-                           C_STRING_WITH_LEN("help_category"),
-                           "help_category", TL_READ);
-  tables[2].init_one_table(C_STRING_WITH_LEN("mysql"),
-                           C_STRING_WITH_LEN("help_relation"),
-                           "help_relation", TL_READ);
-  tables[3].init_one_table(C_STRING_WITH_LEN("mysql"),
-                           C_STRING_WITH_LEN("help_keyword"),
-                           "help_keyword", TL_READ);
+  tables[0].init_one_table(&MYSQL_SCHEMA_NAME, &MYSQL_HELP_TOPIC_NAME, 0, TL_READ);
+  tables[1].init_one_table(&MYSQL_SCHEMA_NAME, &MYSQL_HELP_CATEGORY_NAME, 0, TL_READ);
+  tables[2].init_one_table(&MYSQL_SCHEMA_NAME, &MYSQL_HELP_RELATION_NAME, 0, TL_READ);
+  tables[3].init_one_table(&MYSQL_SCHEMA_NAME, &MYSQL_HELP_KEYWORD_NAME, 0, TL_READ);
   tables[0].next_global= tables[0].next_local= 
     tables[0].next_name_resolution_table= &tables[1];
   tables[1].next_global= tables[1].next_local= 
@@ -719,10 +719,11 @@ static bool mysqld_help_internal(THD *thd, const char *mask)
     Init tables and fields to be usable from items
     tables do not contain VIEWs => we can pass 0 as conds
   */
-  thd->lex->select_lex.context.table_list=
-    thd->lex->select_lex.context.first_name_resolution_table= &tables[0];
-  if (setup_tables(thd, &thd->lex->select_lex.context,
-                   &thd->lex->select_lex.top_join_list,
+  thd->lex->first_select_lex()->context.table_list=
+    thd->lex->first_select_lex()->context.first_name_resolution_table=
+    &tables[0];
+  if (setup_tables(thd, &thd->lex->first_select_lex()->context,
+                   &thd->lex->first_select_lex()->top_join_list,
                    tables, leaves, FALSE, FALSE))
     goto error;
   memcpy((char*) used_fields, (char*) init_used_fields, sizeof(used_fields));
@@ -741,9 +742,6 @@ static bool mysqld_help_internal(THD *thd, const char *mask)
 			      &name, &description, &example);
   delete select;
 
-  if (thd->is_error())
-    goto error;
-
   if (count_topics == 0)
   {
     int UNINIT_VAR(key_id);
