@@ -1,7 +1,7 @@
 /*****************************************************************************
 
 Copyright (c) 1995, 2016, Oracle and/or its affiliates. All Rights Reserved.
-Copyright (c) 2013, 2020, MariaDB Corporation.
+Copyright (c) 2013, 2019, MariaDB Corporation.
 
 This program is free software; you can redistribute it and/or modify it under
 the terms of the GNU General Public License as published by the Free Software
@@ -28,15 +28,15 @@ Created 12/18/1995 Heikki Tuuri
 #define fsp0fsp_h
 
 #include "fsp0types.h"
+#include "fut0lst.h"
+#include "ut0byte.h"
 
 #ifndef UNIV_INNOCHECKSUM
-
-#include "fut0lst.h"
 #include "mtr0mtr.h"
 #include "page0types.h"
 #include "rem0types.h"
-#include "ut0byte.h"
-
+#else
+# include "mach0data.h"
 #endif /* !UNIV_INNOCHECKSUM */
 
 /** @return the PAGE_SSIZE flags for the current innodb_page_size */
@@ -45,6 +45,12 @@ Created 12/18/1995 Heikki Tuuri
 	 0U : (srv_page_size_shift - UNIV_ZIP_SIZE_SHIFT_MIN + 1)	\
 	 << FSP_FLAGS_POS_PAGE_SSIZE)
 
+/** @return the PAGE_SSIZE flags for the current innodb_page_size in
+full checksum format */
+#define FSP_FLAGS_FCRC32_PAGE_SSIZE()					\
+	((srv_page_size_shift - UNIV_ZIP_SIZE_SHIFT_MIN + 1)		\
+	<< FSP_FLAGS_FCRC32_POS_PAGE_SSIZE)
+
 /* @defgroup Compatibility macros for MariaDB 10.1.0 through 10.1.20;
 see the table in fsp0types.h @{ */
 /** Zero relative shift position of the PAGE_COMPRESSION field */
@@ -201,11 +207,6 @@ typedef	byte	fseg_inode_t;
 	(16 + 3 * FLST_BASE_NODE_SIZE			\
 	 + FSEG_FRAG_ARR_N_SLOTS * FSEG_FRAG_SLOT_SIZE)
 
-#define FSP_SEG_INODES_PER_PAGE(page_size)		\
-	((page_size.physical() - FSEG_ARR_OFFSET - 10) / FSEG_INODE_SIZE)
-				/* Number of segment inodes which fit on a
-				single page */
-
 #define FSEG_MAGIC_N_VALUE	97937874
 
 #define	FSEG_FILLFACTOR		8	/* If this value is x, then if
@@ -290,33 +291,6 @@ the extent are free and which contain old tuple version to clean. */
 #ifndef UNIV_INNOCHECKSUM
 /* @} */
 
-/** Calculate the number of pages to extend a datafile.
-We extend single-table tablespaces first one extent at a time,
-but 4 at a time for bigger tablespaces. It is not enough to extend always
-by one extent, because we need to add at least one extent to FSP_FREE.
-A single extent descriptor page will track many extents. And the extent
-that uses its extent descriptor page is put onto the FSP_FREE_FRAG list.
-Extents that do not use their extent descriptor page are added to FSP_FREE.
-The physical page size is used to determine how many extents are tracked
-on one extent descriptor page. See xdes_calc_descriptor_page().
-@param[in]	page_size	page_size of the datafile
-@param[in]	size		current number of pages in the datafile
-@return number of pages to extend the file. */
-ulint
-fsp_get_pages_to_extend_ibd(
-	const page_size_t&	page_size,
-	ulint			size);
-
-/** Calculate the number of physical pages in an extent for this file.
-@param[in]	page_size	page_size of the datafile
-@return number of pages in an extent for this file. */
-UNIV_INLINE
-ulint
-fsp_get_extent_size_in_pages(const page_size_t&	page_size)
-{
-	return (FSP_EXTENT_SIZE << srv_page_size_shift) / page_size.physical();
-}
-
 /**********************************************************************//**
 Reads the space id from the first page of a tablespace.
 @return space id, ULINT UNDEFINED if error */
@@ -347,13 +321,15 @@ fsp_header_get_flags(const page_t* page)
 }
 
 /** Get the byte offset of encryption information in page 0.
-@param[in]	ps	page size
+@param[in]	zip_size	ROW_FORMAT=COMPRESSED page size, or 0
 @return	byte offset relative to FSP_HEADER_OFFSET */
 inline MY_ATTRIBUTE((pure, warn_unused_result))
-ulint
-fsp_header_get_encryption_offset(const page_size_t& ps)
+ulint fsp_header_get_encryption_offset(ulint zip_size)
 {
-	return XDES_ARR_OFFSET + XDES_SIZE * ps.physical() / FSP_EXTENT_SIZE;
+	return zip_size
+		? XDES_ARR_OFFSET + XDES_SIZE * zip_size / FSP_EXTENT_SIZE
+		: XDES_ARR_OFFSET + (XDES_SIZE << srv_page_size_shift)
+		/ FSP_EXTENT_SIZE;
 }
 
 /** Check the encryption key from the first page of a tablespace.
@@ -383,18 +359,26 @@ fsp_header_init_fields(
 void fsp_header_init(fil_space_t* space, ulint size, mtr_t* mtr)
 	MY_ATTRIBUTE((nonnull));
 
-/** Create a new segment.
-@param space                tablespace
-@param byte_offset          byte offset of the created segment header
-@param mtr                  mini-transaction
-@param has_done_reservation whether fsp_reserve_free_extents() was invoked
-@param block                block where segment header is placed,
-                            or NULL to allocate an additional page for that
-@return the block where the segment header is placed, x-latched
-@retval NULL if could not create segment because of lack of space */
+/**********************************************************************//**
+Creates a new segment.
+@return the block where the segment header is placed, x-latched, NULL
+if could not create segment because of lack of space */
 buf_block_t*
-fseg_create(fil_space_t *space, ulint byte_offset, mtr_t *mtr,
-            bool has_done_reservation= false, buf_block_t *block= NULL);
+fseg_create(
+	fil_space_t* space, /*!< in,out: tablespace */
+	ulint	page,	/*!< in: page where the segment header is placed: if
+			this is != 0, the page must belong to another segment,
+			if this is 0, a new page will be allocated and it
+			will belong to the created segment */
+	ulint	byte_offset, /*!< in: byte offset of the created segment header
+			on the page */
+	mtr_t*	mtr,
+   	bool	has_done_reservation = false); /*!< in: whether the caller
+			has already done the reservation for the pages with
+			fsp_reserve_free_extents (at least 2 extents: one for
+			the inode and the other for the segment) then there is
+			no need to do the check for this individual
+			operation */
 
 /**********************************************************************//**
 Calculates the number of pages reserved by a segment, and how many pages are
@@ -504,13 +488,27 @@ fsp_reserve_free_extents(
 @param[in,out]	seg_header	file segment header
 @param[in,out]	space		tablespace
 @param[in]	offset		page number
+@param[in]	ahi		whether we may need to drop the adaptive
+hash index
+@param[in]	log		whether to write MLOG_INIT_FREE_PAGE record
 @param[in,out]	mtr		mini-transaction */
 void
-fseg_free_page(
+fseg_free_page_func(
 	fseg_header_t*	seg_header,
 	fil_space_t*	space,
 	ulint		offset,
+#ifdef BTR_CUR_HASH_ADAPT
+	bool		ahi,
+#endif /* BTR_CUR_HASH_ADAPT */
+	bool		log,
 	mtr_t*		mtr);
+#ifdef BTR_CUR_HASH_ADAPT
+# define fseg_free_page(header, space, offset, ahi, log, mtr)	\
+	fseg_free_page_func(header, space, offset, ahi, log, mtr)
+#else /* BTR_CUR_HASH_ADAPT */
+# define fseg_free_page(header, space, offset, ahi, log, mtr)	\
+	fseg_free_page_func(header, space, offset, log, mtr)
+#endif /* BTR_CUR_HASH_ADAPT */
 /** Determine whether a page is free.
 @param[in,out]	space	tablespace
 @param[in]	page	page number
@@ -523,25 +521,45 @@ Frees part of a segment. This function can be used to free a segment
 by repeatedly calling this function in different mini-transactions.
 Doing the freeing in a single mini-transaction might result in
 too big a mini-transaction.
-@return whether the freeing was completed */
-bool
-fseg_free_step(
+@return TRUE if freeing completed */
+ibool
+fseg_free_step_func(
 	fseg_header_t*	header,	/*!< in, own: segment header; NOTE: if the header
 				resides on the first page of the frag list
 				of the segment, this pointer becomes obsolete
 				after the last freeing step */
+#ifdef BTR_CUR_HASH_ADAPT
+	bool		ahi,	/*!< in: whether we may need to drop
+				the adaptive hash index */
+#endif /* BTR_CUR_HASH_ADAPT */
 	mtr_t*		mtr)	/*!< in/out: mini-transaction */
 	MY_ATTRIBUTE((warn_unused_result));
+#ifdef BTR_CUR_HASH_ADAPT
+# define fseg_free_step(header, ahi, mtr) fseg_free_step_func(header, ahi, mtr)
+#else /* BTR_CUR_HASH_ADAPT */
+# define fseg_free_step(header, ahi, mtr) fseg_free_step_func(header, mtr)
+#endif /* BTR_CUR_HASH_ADAPT */
 /**********************************************************************//**
 Frees part of a segment. Differs from fseg_free_step because this function
 leaves the header page unfreed.
-@return whether the freeing was completed, except for the header page */
-bool
-fseg_free_step_not_header(
+@return TRUE if freeing completed, except the header page */
+ibool
+fseg_free_step_not_header_func(
 	fseg_header_t*	header,	/*!< in: segment header which must reside on
 				the first fragment page of the segment */
+#ifdef BTR_CUR_HASH_ADAPT
+	bool		ahi,	/*!< in: whether we may need to drop
+				the adaptive hash index */
+#endif /* BTR_CUR_HASH_ADAPT */
 	mtr_t*		mtr)	/*!< in/out: mini-transaction */
 	MY_ATTRIBUTE((warn_unused_result));
+#ifdef BTR_CUR_HASH_ADAPT
+# define fseg_free_step_not_header(header, ahi, mtr)	\
+	fseg_free_step_not_header_func(header, ahi, mtr)
+#else /* BTR_CUR_HASH_ADAPT */
+# define fseg_free_step_not_header(header, ahi, mtr)	\
+	fseg_free_step_not_header_func(header, mtr)
+#endif /* BTR_CUR_HASH_ADAPT */
 
 /** Reset the page type.
 Data files created before MySQL 5.1.48 may contain garbage in FIL_PAGE_TYPE.
@@ -583,13 +601,12 @@ fil_block_check_type(
 
 /** Checks if a page address is an extent descriptor page address.
 @param[in]	page_id		page id
-@param[in]	page_size	page size
-@return TRUE if a descriptor page */
-UNIV_INLINE
-ibool
-fsp_descr_page(
-	const page_id_t		page_id,
-	const page_size_t&	page_size);
+@param[in]	physical_size	page size
+@return whether a descriptor page */
+inline bool fsp_descr_page(const page_id_t page_id, ulint physical_size)
+{
+	return (page_id.page_no() & (physical_size - 1)) == FSP_XDES_OFFSET;
+}
 
 /** Initialize a file page whose prior contents should be ignored.
 @param[in,out]	block	buffer pool block */
@@ -636,7 +653,7 @@ fsp_flags_convert_from_101(ulint flags)
 {
 	DBUG_EXECUTE_IF("fsp_flags_is_valid_failure",
 			return(ULINT_UNDEFINED););
-	if (flags == 0) {
+	if (flags == 0 || fil_space_t::full_crc32(flags)) {
 		return(flags);
 	}
 
@@ -731,7 +748,7 @@ fsp_flags_convert_from_101(ulint flags)
 	flags = ((flags & 0x3f) | ssize << FSP_FLAGS_POS_PAGE_SSIZE
 		 | FSP_FLAGS_GET_PAGE_COMPRESSION_MARIADB101(flags)
 		 << FSP_FLAGS_POS_PAGE_COMPRESSION);
-	ut_ad(fsp_flags_is_valid(flags, false));
+	ut_ad(fil_space_t::is_valid_flags(flags, false));
 	return(flags);
 }
 
@@ -745,7 +762,7 @@ bool
 fsp_flags_match(ulint expected, ulint actual)
 {
 	expected &= ~FSP_FLAGS_MEM_MASK;
-	ut_ad(fsp_flags_is_valid(expected, false));
+	ut_ad(fil_space_t::is_valid_flags(expected, false));
 
 	if (actual == expected) {
 		return(true);
@@ -755,16 +772,6 @@ fsp_flags_match(ulint expected, ulint actual)
 	return(actual == expected);
 }
 
-/** Calculates the descriptor index within a descriptor page.
-@param[in]	page_size	page size
-@param[in]	offset		page offset
-@return descriptor index */
-UNIV_INLINE
-ulint
-xdes_calc_descriptor_index(
-	const page_size_t&	page_size,
-	ulint			offset);
-
 /**********************************************************************//**
 Gets a descriptor bit of a page.
 @return TRUE if free */
@@ -777,15 +784,42 @@ xdes_get_bit(
 	ulint		offset);/*!< in: page offset within extent:
 				0 ... FSP_EXTENT_SIZE - 1 */
 
-/** Calculates the page where the descriptor of a page resides.
-@param[in]	page_size	page size
+/** Determine the descriptor index within a descriptor page.
+@param[in]	zip_size	ROW_FORMAT=COMPRESSED page size, or 0
+@param[in]	offset		page offset
+@return descriptor index */
+inline ulint xdes_calc_descriptor_index(ulint zip_size, ulint offset)
+{
+	return ut_2pow_remainder<ulint>(offset,
+					zip_size ? zip_size : srv_page_size)
+		/ FSP_EXTENT_SIZE;
+}
+
+/** Determine the descriptor page number for a page.
+@param[in]	zip_size	ROW_FORMAT=COMPRESSED page size, or 0
 @param[in]	offset		page offset
 @return descriptor page offset */
-UNIV_INLINE
-ulint
-xdes_calc_descriptor_page(
-	const page_size_t&	page_size,
-	ulint			offset);
+inline ulint xdes_calc_descriptor_page(ulint zip_size, ulint offset)
+{
+	compile_time_assert(UNIV_PAGE_SIZE_MAX > XDES_ARR_OFFSET
+			    + (UNIV_PAGE_SIZE_MAX / FSP_EXTENT_SIZE_MAX)
+			    * XDES_SIZE_MAX);
+	compile_time_assert(UNIV_PAGE_SIZE_MIN > XDES_ARR_OFFSET
+			    + (UNIV_PAGE_SIZE_MIN / FSP_EXTENT_SIZE_MIN)
+			    * XDES_SIZE_MIN);
+
+	ut_ad(srv_page_size > XDES_ARR_OFFSET
+	      + (srv_page_size / FSP_EXTENT_SIZE)
+	      * XDES_SIZE);
+	ut_ad(UNIV_ZIP_SIZE_MIN > XDES_ARR_OFFSET
+	      + (UNIV_ZIP_SIZE_MIN / FSP_EXTENT_SIZE)
+	      * XDES_SIZE);
+	ut_ad(!zip_size
+	      || zip_size > XDES_ARR_OFFSET
+	      + (zip_size / FSP_EXTENT_SIZE) * XDES_SIZE);
+	return ut_2pow_round<ulint>(offset,
+				    zip_size ? zip_size : srv_page_size);
+}
 
 #endif /* UNIV_INNOCHECKSUM */
 
