@@ -96,6 +96,10 @@ those defined in mysql file ft_global.h */
 /** Threshold where our optimize thread automatically kicks in */
 #define FTS_OPTIMIZE_THRESHOLD		10000000
 
+/** Threshold to avoid exhausting of doc ids. Consecutive doc id difference
+should not exceed FTS_DOC_ID_MAX_STEP */
+#define FTS_DOC_ID_MAX_STEP		65535
+
 /** Maximum possible Fulltext word length in bytes (assuming mbmaxlen=4) */
 #define FTS_MAX_WORD_LEN		(HA_FT_MAXCHARLEN * 4)
 
@@ -313,11 +317,19 @@ public:
 	/** fts_t destructor. */
 	~fts_t();
 
-	/** Whether the ADDED table record sync-ed after crash recovery */
+	/** Mutex protecting bg_threads* and fts_add_wq. */
+	ib_mutex_t	bg_threads_mutex;
+
+	/** Whether the ADDED table record sync-ed after
+	crash recovery; protected by bg_threads_mutex */
 	unsigned	added_synced:1;
-	/** Whether the table holds dict_sys->mutex */
+	/** Whether the table holds dict_sys.mutex;
+	protected by bg_threads_mutex */
 	unsigned	dict_locked:1;
 
+	/** Number of background threads accessing this table. */
+	ulint		bg_threads;
+
 	/** Work queue for scheduling jobs for the FTS 'Add' thread, or NULL
 	if the thread has not yet been created. Each work item is a
 	fts_trx_doc_ids_t*. */
@@ -372,9 +384,9 @@ extern bool		fts_need_sync;
 
 #define	fts_que_graph_free(graph)			\
 do {							\
-	mutex_enter(&dict_sys->mutex);			\
+	mutex_enter(&dict_sys.mutex);			\
 	que_graph_free(graph);				\
-	mutex_exit(&dict_sys->mutex);			\
+	mutex_exit(&dict_sys.mutex);			\
 } while (0)
 
 /******************************************************************//**
@@ -572,17 +584,15 @@ fts_get_doc_id_from_row(
 						want to extract.*/
 
 /** Extract the doc id from the record that belongs to index.
-@param[in]	table	table
-@param[in]	rec	record contains FTS_DOC_ID
+@param[in]	rec	record containing FTS_DOC_ID
 @param[in]	index	index of rec
-@param[in]	heap	heap memory
+@param[in]	offsets	rec_get_offsets(rec,index)
 @return doc id that was extracted from rec */
 doc_id_t
 fts_get_doc_id_from_rec(
-        dict_table_t*           table,
-        const rec_t*            rec,
-        const dict_index_t*     index,
-        mem_heap_t*             heap);
+	const rec_t*		rec,
+	const dict_index_t*	index,
+	const rec_offs*		offsets);
 
 /** Add new fts doc id to the update vector.
 @param[in]	table		the table that contains the FTS index.
@@ -724,9 +734,12 @@ fts_savepoint_rollback_last_stmt(
 /*=============================*/
 	trx_t*		trx);			/*!< in: transaction */
 
-/** Drop all orphaned FTS auxiliary tables, those that don't have a parent
+/***********************************************************************//**
+Drop all orphaned FTS auxiliary tables, those that don't have a parent
 table or FTS index defined on them. */
-void fts_drop_orphaned_tables();
+void
+fts_drop_orphaned_tables(void);
+/*==========================*/
 
 /** Run SYNC on the table, i.e., write out data from the cache to the
 FTS auxiliary INDEX table and clear the cache at the end.
@@ -736,7 +749,7 @@ FTS auxiliary INDEX table and clear the cache at the end.
 dberr_t fts_sync_table(dict_table_t* table, bool wait = true);
 
 /****************************************************************//**
-Free the query graph but check whether dict_sys->mutex is already
+Free the query graph but check whether dict_sys.mutex is already
 held */
 void
 fts_que_graph_free_check_lock(
@@ -760,6 +773,15 @@ fts_init_doc_id(
 /*============*/
 	const dict_table_t*		table);	/*!< in: table */
 
+/* Get parent table name if it's a fts aux table
+@param[in]	aux_table_name	aux table name
+@param[in]	aux_table_len	aux table length
+@return parent table name, or NULL */
+char*
+fts_get_parent_table_name(
+	const char*	aux_table_name,
+	ulint		aux_table_len);
+
 /******************************************************************//**
 compare two character string according to their charset. */
 extern
@@ -960,20 +982,4 @@ fts_trx_t*
 fts_trx_create(
 	trx_t*  trx);
 
-/** Clear all fts resources when there is no internal DOC_ID
-and there are no new fts index to add.
-@param[in,out]  table   table  where fts is to be freed
-@param[in]      trx     transaction to drop all fts tables */
-void fts_clear_all(dict_table_t *table, trx_t *trx);
-
-/** Check whether the given name is fts auxiliary table
-and fetch the parent table id and index id
-@param[in]	name		table name
-@param[in,out]	table_id	parent table id
-@param[in,out]	index_id	index id
-@return true if it is auxilary table */
-bool fts_check_aux_table(const char *name,
-                         table_id_t *table_id,
-                         index_id_t *index_id);
-
 #endif /*!< fts0fts.h */
