@@ -2,7 +2,7 @@
 
 Copyright (c) 1995, 2016, Oracle and/or its affiliates. All Rights Reserved.
 Copyright (c) 2008, Google Inc.
-Copyright (c) 2017, 2020, MariaDB Corporation.
+Copyright (c) 2017, 2019, MariaDB Corporation.
 
 Portions of this file contain modifications contributed and copyrighted by
 Google, Inc. Those modifications are gratefully acknowledged and are described
@@ -499,16 +499,16 @@ bool
 rw_lock_lock_word_decr(
 /*===================*/
 	rw_lock_t*	lock,		/*!< in/out: rw-lock */
-	ulint		amount,		/*!< in: amount to decrement */
-	lint		threshold);	/*!< in: threshold of judgement */
+	int32_t		amount,		/*!< in: amount to decrement */
+	int32_t		threshold);	/*!< in: threshold of judgement */
 #ifdef UNIV_DEBUG
 /******************************************************************//**
 Checks if the thread has locked the rw-lock in the specified mode, with
 the pass value == 0. */
-ibool
+bool
 rw_lock_own(
 /*========*/
-	const rw_lock_t*lock,		/*!< in: rw-lock */
+	rw_lock_t*	lock,		/*!< in: rw-lock */
 	ulint		lock_type)	/*!< in: lock type: RW_LOCK_S,
 					RW_LOCK_X */
 	MY_ATTRIBUTE((warn_unused_result));
@@ -569,14 +569,20 @@ struct rw_lock_t
 #endif /* UNIV_DEBUG */
 {
 	/** Holds the state of the lock. */
-	volatile lint	lock_word;
+	std::atomic<int32_t>	lock_word;
 
 	/** 1: there are waiters */
-	volatile uint32_t	waiters;
+	std::atomic<int32_t>	waiters;
 
 	/** number of granted SX locks. */
 	volatile ulint	sx_recursive;
 
+	/** This is TRUE if the writer field is RW_LOCK_X_WAIT; this field
+	is located far from the memory update hotspot fields which are at
+	the start of this struct, thus we can peek this field without
+	causing much memory bus traffic */
+	bool		writer_is_wait_ex;
+
 	/** The value is typically set to thread id of a writer thread making
 	normal rw_locks recursive. In case of asynchronous IO, when a non-zero
 	value of 'pass' is passed then we keep the lock non-recursive.
@@ -595,9 +601,6 @@ struct rw_lock_t
 	/** File name where lock created */
 	const char*	cfile_name;
 
-	/** last s-lock file/line is not guaranteed to be correct */
-	const char*	last_s_file_name;
-
 	/** File name where last x-locked */
 	const char*	last_x_file_name;
 
@@ -607,9 +610,6 @@ struct rw_lock_t
 	/** If 1 then the rw-lock is a block lock */
 	unsigned	is_block_lock:1;
 
-	/** Line number where last time s-locked */
-	unsigned	last_s_line:14;
-
 	/** Line number where last time x-locked */
 	unsigned	last_x_line:14;
 
@@ -625,8 +625,7 @@ struct rw_lock_t
 #endif /* UNIV_PFS_RWLOCK */
 
 #ifdef UNIV_DEBUG
-	virtual std::string to_string() const;
-	virtual std::string locked_from() const;
+	std::string to_string() const override;
 
 	/** In the debug version: pointer to the debug info list of the lock */
 	UT_LIST_BASE_NODE_T(rw_lock_debug_t) debug_list;
@@ -634,7 +633,6 @@ struct rw_lock_t
 	/** Level in the global latching order. */
 	latch_level_t	level;
 #endif /* UNIV_DEBUG */
-
 };
 #ifdef UNIV_DEBUG
 /** The structure for storing debug info of an rw-lock.  All access to this
