@@ -1,7 +1,7 @@
 /*****************************************************************************
 
 Copyright (c) 1996, 2016, Oracle and/or its affiliates. All Rights Reserved.
-Copyright (c) 2015, 2021, MariaDB Corporation.
+Copyright (c) 2015, 2020, MariaDB Corporation.
 
 This program is free software; you can redistribute it and/or modify it under
 the terms of the GNU General Public License as published by the Free Software
@@ -112,7 +112,7 @@ trx_init(
 
 	trx->op_info = "";
 
-	trx->active_commit_ordered = false;
+	trx->active_commit_ordered = 0;
 
 	trx->isolation_level = TRX_ISO_REPEATABLE_READ;
 
@@ -161,11 +161,6 @@ trx_init(
 	trx->lock.rec_cached = 0;
 
 	trx->lock.table_cached = 0;
-#ifdef WITH_WSREP
-	ut_ad(!trx->wsrep);
-	ut_ad(!trx->wsrep_event);
-	ut_ad(!trx->wsrep_UK_scan);
-#endif /* WITH_WSREP */
 
 	ut_ad(trx->get_flush_observer() == NULL);
 }
@@ -205,6 +200,9 @@ struct TrxFactory {
 
 		lock_trx_lock_list_init(&trx->lock.trx_locks);
 
+		UT_LIST_INIT(trx->lock.evicted_tables,
+			     &dict_table_t::table_LRU);
+
 		UT_LIST_INIT(
 			trx->trx_savepoints,
 			&trx_named_savept_t::trx_savepoints);
@@ -216,15 +214,6 @@ struct TrxFactory {
 	@param trx the transaction for which to release resources */
 	static void destroy(trx_t* trx)
 	{
-#ifdef __SANITIZE_ADDRESS__
-		/* Unpoison the memory for AddressSanitizer */
-		MEM_MAKE_ADDRESSABLE(trx, sizeof *trx);
-#else
-		/* Declare the contents as initialized for Valgrind;
-		we checked this in trx_t::free(). */
-		MEM_MAKE_DEFINED(trx, sizeof *trx);
-#endif
-
 		ut_a(trx->magic_n == TRX_MAGIC_N);
 		ut_ad(!trx->mysql_thd);
 
@@ -238,6 +227,7 @@ struct TrxFactory {
 		}
 
 		ut_a(UT_LIST_GET_LEN(trx->lock.trx_locks) == 0);
+		ut_ad(UT_LIST_GET_LEN(trx->lock.evicted_tables) == 0);
 
 		UT_DELETE(trx->xid);
 		ut_free(trx->detailed_error);
@@ -252,6 +242,36 @@ struct TrxFactory {
 
 		trx->read_view.~ReadView();
 	}
+
+	/** Enforce any invariants here, this is called before the transaction
+	is added to the pool.
+	@return true if all OK */
+	static bool debug(const trx_t* trx)
+	{
+		ut_a(trx->error_state == DB_SUCCESS);
+
+		ut_a(trx->magic_n == TRX_MAGIC_N);
+
+		ut_ad(!trx->read_only);
+
+		ut_ad(trx->state == TRX_STATE_NOT_STARTED);
+
+		ut_ad(trx->dict_operation == TRX_DICT_OP_NONE);
+
+		ut_ad(trx->mysql_thd == 0);
+
+		ut_a(trx->lock.wait_thr == NULL);
+		ut_a(trx->lock.wait_lock == NULL);
+		ut_a(trx->dict_operation_lock_mode == 0);
+
+		ut_a(UT_LIST_GET_LEN(trx->lock.trx_locks) == 0);
+
+		ut_ad(trx->autoinc_locks == NULL);
+
+		ut_ad(trx->lock.table_locks.empty());
+
+		return(true);
+	}
 };
 
 /** The lock strategy for TrxPool */
@@ -328,23 +348,11 @@ trx_pool_close()
 	trx_pools = 0;
 }
 
-/** @return an allocated transaction */
+/** @return a trx_t instance from trx_pools. */
 trx_t *trx_create()
 {
 	trx_t*	trx = trx_pools->get();
 
-#ifdef __SANITIZE_ADDRESS__
-	/* Unpoison the memory for AddressSanitizer.
-	It may have been poisoned in trx_t::free().*/
-	MEM_MAKE_ADDRESSABLE(trx, sizeof *trx);
-#else
-	/* Declare the memory initialized for Valgrind.
-	The trx_t that are released to the pool are
-	actually initialized; we checked that by
-	MEM_CHECK_DEFINED() in trx_t::free(). */
-	MEM_MAKE_DEFINED(trx, sizeof *trx);
-#endif
-
 	assert_trx_is_free(trx);
 
 	mem_heap_t*	heap;
@@ -361,16 +369,21 @@ trx_t *trx_create()
 
 	alloc = ib_heap_allocator_create(heap);
 
+	/* Remember to free the vector explicitly in trx_free(). */
 	trx->autoinc_locks = ib_vector_create(alloc, sizeof(void**), 4);
 
+	/* Should have been either just initialized or .clear()ed by
+	trx_free(). */
 	ut_ad(trx->mod_tables.empty());
+	ut_ad(trx->lock.table_locks.empty());
+	ut_ad(UT_LIST_GET_LEN(trx->lock.trx_locks) == 0);
 	ut_ad(trx->lock.n_rec_locks == 0);
 	ut_ad(trx->lock.table_cached == 0);
 	ut_ad(trx->lock.rec_cached == 0);
+	ut_ad(UT_LIST_GET_LEN(trx->lock.evicted_tables) == 0);
 
 #ifdef WITH_WSREP
 	trx->wsrep_event = NULL;
-	ut_ad(!trx->wsrep_UK_scan);
 #endif /* WITH_WSREP */
 
 	trx_sys.register_trx(trx);
@@ -378,110 +391,79 @@ trx_t *trx_create()
 	return(trx);
 }
 
-/** Free the memory to trx_pools */
-void trx_t::free()
+/**
+  Release a trx_t instance back to the pool.
+  @param trx the instance to release.
+*/
+void trx_free(trx_t*& trx)
 {
-  MEM_CHECK_DEFINED(this, sizeof *this);
+	ut_ad(!trx->declared_to_be_inside_innodb);
+	ut_ad(!trx->n_mysql_tables_in_use);
+	ut_ad(!trx->mysql_n_tables_locked);
+	ut_ad(!trx->internal);
 
-  ut_ad(!n_mysql_tables_in_use);
-  ut_ad(!mysql_n_tables_locked);
-  ut_ad(!internal);
-  ut_ad(!declared_to_be_inside_innodb);
-  ut_ad(!will_lock);
-  ut_ad(error_state == DB_SUCCESS);
-  ut_ad(magic_n == TRX_MAGIC_N);
-  ut_ad(!read_only);
-  ut_ad(!lock.wait_lock);
+	if (trx->declared_to_be_inside_innodb) {
 
-  dict_operation= TRX_DICT_OP_NONE;
-  trx_sys.deregister_trx(this);
-  assert_trx_is_free(this);
-  trx_sys.rw_trx_hash.put_pins(this);
+		ib::error() << "Freeing a trx (" << trx << ", "
+			<< trx_get_id_for_print(trx) << ") which is declared"
+			" to be processing inside InnoDB";
 
-  mysql_thd= NULL;
-  mysql_log_file_name= NULL;
+		trx_print(stderr, trx, 600);
+		putc('\n', stderr);
 
-  // FIXME: We need to avoid this heap free/alloc for each commit.
-  if (autoinc_locks)
-  {
-    ut_ad(ib_vector_is_empty(autoinc_locks));
-    /* We allocated a dedicated heap for the vector. */
-    ib_vector_free(autoinc_locks);
-    autoinc_locks= NULL;
-  }
+		/* This is an error but not a fatal error. We must keep
+		the counters like srv_conc.n_active accurate. */
+		srv_conc_force_exit_innodb(trx);
+	}
 
-  mod_tables.clear();
+	if (trx->n_mysql_tables_in_use != 0
+	    || trx->mysql_n_tables_locked != 0) {
 
-  MEM_NOACCESS(&n_ref, sizeof n_ref);
-  /* do not poison mutex */
-  MEM_NOACCESS(&id, sizeof id);
-  MEM_NOACCESS(&no, sizeof no);
-  MEM_NOACCESS(&state, sizeof state);
-  MEM_NOACCESS(&is_recovered, sizeof is_recovered);
-#ifdef WITH_WSREP
-  MEM_NOACCESS(&wsrep, sizeof wsrep);
-#endif
-  MEM_NOACCESS(&read_view, sizeof read_view);
-  MEM_NOACCESS(&trx_list, sizeof trx_list);
-  MEM_NOACCESS(&lock, sizeof lock);
-  MEM_NOACCESS(&op_info, sizeof op_info);
-  MEM_NOACCESS(&isolation_level, sizeof isolation_level);
-  MEM_NOACCESS(&check_foreigns, sizeof check_foreigns);
-  MEM_NOACCESS(&is_registered, sizeof is_registered);
-  MEM_NOACCESS(&active_commit_ordered, sizeof active_commit_ordered);
-  MEM_NOACCESS(&check_unique_secondary, sizeof check_unique_secondary);
-  MEM_NOACCESS(&flush_log_later, sizeof flush_log_later);
-  MEM_NOACCESS(&must_flush_log_later, sizeof must_flush_log_later);
-  MEM_NOACCESS(&duplicates, sizeof duplicates);
-  MEM_NOACCESS(&dict_operation, sizeof dict_operation);
-  MEM_NOACCESS(&declared_to_be_inside_innodb, sizeof declared_to_be_inside_innodb);
-  MEM_NOACCESS(&n_tickets_to_enter_innodb, sizeof n_tickets_to_enter_innodb);
-  MEM_NOACCESS(&dict_operation_lock_mode, sizeof dict_operation_lock_mode);
-  MEM_NOACCESS(&start_time, sizeof start_time);
-  MEM_NOACCESS(&start_time_micro, sizeof start_time_micro);
-  MEM_NOACCESS(&commit_lsn, sizeof commit_lsn);
-  MEM_NOACCESS(&table_id, sizeof table_id);
-  MEM_NOACCESS(&mysql_thd, sizeof mysql_thd);
-  MEM_NOACCESS(&mysql_log_file_name, sizeof mysql_log_file_name);
-  MEM_NOACCESS(&mysql_log_offset, sizeof mysql_log_offset);
-  MEM_NOACCESS(&n_mysql_tables_in_use, sizeof n_mysql_tables_in_use);
-  MEM_NOACCESS(&mysql_n_tables_locked, sizeof mysql_n_tables_locked);
-  MEM_NOACCESS(&error_state, sizeof error_state);
-  MEM_NOACCESS(&error_info, sizeof error_info);
-  MEM_NOACCESS(&error_key_num, sizeof error_key_num);
-  MEM_NOACCESS(&graph, sizeof graph);
-  MEM_NOACCESS(&trx_savepoints, sizeof trx_savepoints);
-  MEM_NOACCESS(&undo_no, sizeof undo_no);
-  MEM_NOACCESS(&last_sql_stat_start, sizeof last_sql_stat_start);
-  MEM_NOACCESS(&rsegs, sizeof rsegs);
-  MEM_NOACCESS(&roll_limit, sizeof roll_limit);
-  MEM_NOACCESS(&in_rollback, sizeof in_rollback);
-  MEM_NOACCESS(&pages_undone, sizeof pages_undone);
-  MEM_NOACCESS(&n_autoinc_rows, sizeof n_autoinc_rows);
-  MEM_NOACCESS(&autoinc_locks, sizeof autoinc_locks);
-  MEM_NOACCESS(&read_only, sizeof read_only);
-  MEM_NOACCESS(&auto_commit, sizeof auto_commit);
-  MEM_NOACCESS(&will_lock, sizeof will_lock);
-  MEM_NOACCESS(&fts_trx, sizeof fts_trx);
-  MEM_NOACCESS(&fts_next_doc_id, sizeof fts_next_doc_id);
-  MEM_NOACCESS(&flush_tables, sizeof flush_tables);
-  MEM_NOACCESS(&ddl, sizeof ddl);
-  MEM_NOACCESS(&internal, sizeof internal);
-#ifdef UNIV_DEBUG
-  MEM_NOACCESS(&start_line, sizeof start_line);
-  MEM_NOACCESS(&start_file, sizeof start_file);
-#endif /* UNIV_DEBUG */
-  MEM_NOACCESS(&xid, sizeof xid);
-  MEM_NOACCESS(&mod_tables, sizeof mod_tables);
-  MEM_NOACCESS(&detailed_error, sizeof detailed_error);
-  MEM_NOACCESS(&flush_observer, sizeof flush_observer);
-#ifdef WITH_WSREP
-  MEM_NOACCESS(&wsrep_event, sizeof wsrep_event);
-  ut_ad(!wsrep_UK_scan);
-  MEM_NOACCESS(&wsrep_UK_scan, sizeof wsrep_UK_scan);
-#endif /* WITH_WSREP */
-  MEM_NOACCESS(&magic_n, sizeof magic_n);
-  trx_pools->mem_free(this);
+		ib::error() << "MySQL is freeing a thd though"
+			" trx->n_mysql_tables_in_use is "
+			<< trx->n_mysql_tables_in_use
+			<< " and trx->mysql_n_tables_locked is "
+			<< trx->mysql_n_tables_locked << ".";
+
+		trx_print(stderr, trx, 600);
+		ut_print_buf(stderr, trx, sizeof(trx_t));
+		putc('\n', stderr);
+	}
+
+	trx->dict_operation = TRX_DICT_OP_NONE;
+	assert_trx_is_inactive(trx);
+
+	trx_sys.deregister_trx(trx);
+
+	assert_trx_is_free(trx);
+
+	trx_sys.rw_trx_hash.put_pins(trx);
+	trx->mysql_thd = 0;
+	trx->mysql_log_file_name = 0;
+
+	// FIXME: We need to avoid this heap free/alloc for each commit.
+	if (trx->autoinc_locks != NULL) {
+		ut_ad(ib_vector_is_empty(trx->autoinc_locks));
+		/* We allocated a dedicated heap for the vector. */
+		ib_vector_free(trx->autoinc_locks);
+		trx->autoinc_locks = NULL;
+	}
+
+	trx->mod_tables.clear();
+
+	/* trx locking state should have been reset before returning trx
+	to pool */
+	ut_ad(trx->will_lock == 0);
+
+	trx_pools->mem_free(trx);
+	/* Unpoison the memory for innodb_monitor_set_option;
+	it is operating also on the freed transaction objects. */
+	MEM_UNDEFINED(&trx->mutex, sizeof trx->mutex);
+	/* Declare the contents as initialized for Valgrind;
+	we checked that it was initialized in trx_pools->mem_free(trx). */
+	UNIV_MEM_VALID(&trx->mutex, sizeof trx->mutex);
+
+	trx = NULL;
 }
 
 /** Transition to committed state, to release implicit locks. */
@@ -551,7 +533,8 @@ trx_free_at_shutdown(trx_t *trx)
 	trx->state = TRX_STATE_NOT_STARTED;
 	ut_ad(!UT_LIST_GET_LEN(trx->lock.trx_locks));
 	trx->id = 0;
-	trx->free();
+
+	trx_free(trx);
 }
 
 
@@ -633,10 +616,10 @@ trx_resurrect_table_locks(
 		if (dict_table_t* table = dict_table_open_on_id(
 			    *i, FALSE, DICT_TABLE_OP_LOAD_TABLESPACE)) {
 			if (!table->is_readable()) {
-				mutex_enter(&dict_sys->mutex);
+				mutex_enter(&dict_sys.mutex);
 				dict_table_close(table, TRUE, FALSE);
-				dict_table_remove_from_cache(table);
-				mutex_exit(&dict_sys->mutex);
+				dict_sys.remove(table);
+				mutex_exit(&dict_sys.mutex);
 				continue;
 			}
 
@@ -1277,16 +1260,10 @@ trx_update_mod_tables_timestamp(
 
 	trx_mod_tables_t::const_iterator	end = trx->mod_tables.end();
 #ifdef UNIV_DEBUG
-# if MYSQL_VERSION_ID >= 100405
-#  define dict_sys_mutex dict_sys.mutex
-# else
-#  define dict_sys_mutex dict_sys->mutex
-# endif
-
 	const bool preserve_tables = !innodb_evict_tables_on_commit_debug
 		|| trx->is_recovered /* avoid trouble with XA recovery */
 # if 1 /* if dict_stats_exec_sql() were not playing dirty tricks */
-		|| mutex_own(&dict_sys_mutex)
+		|| mutex_own(&dict_sys.mutex)
 # else /* this would be more proper way to do it */
 		|| trx->dict_operation_lock_mode || trx->dict_operation
 # endif
@@ -1308,8 +1285,7 @@ trx_update_mod_tables_timestamp(
 		dict_table_t* table = it->first;
 		table->update_time = now;
 #ifdef UNIV_DEBUG
-		if (preserve_tables || table->get_ref_count()
-		    || UT_LIST_GET_LEN(table->locks)) {
+		if (preserve_tables || table->get_ref_count()) {
 			/* do not evict when committing DDL operations
 			or if some other transaction is holding the
 			table handle */
@@ -1317,320 +1293,302 @@ trx_update_mod_tables_timestamp(
 		}
 		/* recheck while holding the mutex that blocks
 		table->acquire() */
-		mutex_enter(&dict_sys_mutex);
-		mutex_enter(&lock_sys.mutex);
-		const bool do_evict = !table->get_ref_count()
-			&& !UT_LIST_GET_LEN(table->locks);
-		mutex_exit(&lock_sys.mutex);
-		if (do_evict) {
-# if MYSQL_VERSION_ID >= 100405
+		mutex_enter(&dict_sys.mutex);
+		if (!table->get_ref_count()) {
 			dict_sys.remove(table, true);
-# else
-			dict_table_remove_from_cache_low(table, true);
-# endif
 		}
-		mutex_exit(&dict_sys_mutex);
+		mutex_exit(&dict_sys.mutex);
 #endif
 	}
 
 	trx->mod_tables.clear();
 }
 
-/****************************************************************//**
-Commits a transaction in memory. */
-static
-void
-trx_commit_in_memory(
-/*=================*/
-	trx_t*		trx,	/*!< in/out: transaction */
-	const mtr_t*	mtr)	/*!< in: mini-transaction of
-				trx_write_serialisation_history(), or NULL if
-				the transaction did not modify anything */
+/** Evict a table definition due to the rollback of ALTER TABLE.
+@param[in]	table_id	table identifier */
+void trx_t::evict_table(table_id_t table_id)
 {
-	trx->must_flush_log_later = false;
-	trx->read_view.close();
+	ut_ad(in_rollback);
 
-	if (trx_is_autocommit_non_locking(trx)) {
-		ut_ad(trx->id == 0);
-		ut_ad(trx->read_only);
-		ut_a(!trx->is_recovered);
-		ut_ad(trx->rsegs.m_redo.rseg == NULL);
+	dict_table_t* table = dict_table_open_on_id(
+		table_id, true, DICT_TABLE_OP_OPEN_ONLY_IF_CACHED);
+	if (!table) {
+		return;
+	}
 
-		/* Note: We are asserting without holding the lock mutex. But
-		that is OK because this transaction is not waiting and cannot
-		be rolled back and no new locks can (or should) be added
-		because it is flagged as a non-locking read-only transaction. */
+	if (!table->release()) {
+		/* This must be a DDL operation that is being rolled
+		back in an active connection. */
+		ut_a(table->get_ref_count() == 1);
+		ut_ad(!is_recovered);
+		ut_ad(mysql_thd);
+		return;
+	}
 
-		ut_a(UT_LIST_GET_LEN(trx->lock.trx_locks) == 0);
+	/* This table should only be locked by this transaction, if at all. */
+	ut_ad(UT_LIST_GET_LEN(table->locks) <= 1);
+	const bool locked = UT_LIST_GET_LEN(table->locks);
+	ut_ad(!locked || UT_LIST_GET_FIRST(table->locks)->trx == this);
+	dict_sys.remove(table, true, locked);
+	if (locked) {
+		UT_LIST_ADD_FIRST(lock.evicted_tables, table);
+	}
+}
 
-		/* This state change is not protected by any mutex, therefore
-		there is an inherent race here around state transition during
-		printouts. We ignore this race for the sake of efficiency.
-		However, the trx_sys_t::mutex will protect the trx_t instance
-		and it cannot be removed from the trx_list and freed
-		without first acquiring the trx_sys_t::mutex. */
+/** Mark a transaction committed in the main memory data structures. */
+inline void trx_t::commit_in_memory(const mtr_t *mtr)
+{
+  must_flush_log_later= false;
+  read_view.close();
 
-		ut_ad(trx_state_eq(trx, TRX_STATE_ACTIVE));
+  if (trx_is_autocommit_non_locking(this))
+  {
+    ut_ad(id == 0);
+    ut_ad(read_only);
+    ut_a(!is_recovered);
+    ut_ad(!rsegs.m_redo.rseg);
 
-		MONITOR_INC(MONITOR_TRX_NL_RO_COMMIT);
+    /* Note: We are asserting without holding the lock mutex. But
+    that is OK because this transaction is not waiting and cannot
+    be rolled back and no new locks can (or should) be added
+    because it is flagged as a non-locking read-only transaction. */
+    ut_a(UT_LIST_GET_LEN(lock.trx_locks) == 0);
 
-		DBUG_LOG("trx", "Autocommit in memory: " << trx);
-		trx->state = TRX_STATE_NOT_STARTED;
-	} else {
+    /* This state change is not protected by any mutex, therefore
+    there is an inherent race here around state transition during
+    printouts. We ignore this race for the sake of efficiency.
+    However, the trx_sys_t::mutex will protect the trx_t instance
+    and it cannot be removed from the trx_list and freed
+    without first acquiring the trx_sys_t::mutex. */
+    ut_ad(trx_state_eq(this, TRX_STATE_ACTIVE));
+
+    MONITOR_INC(MONITOR_TRX_NL_RO_COMMIT);
+
+    DBUG_LOG("trx", "Autocommit in memory: " << this);
+    state= TRX_STATE_NOT_STARTED;
+  }
+  else
+  {
 #ifdef UNIV_DEBUG
-		if (!UT_LIST_GET_LEN(trx->lock.trx_locks)) {
-			for (lock_list::iterator it
-				= trx->lock.table_locks.begin();
-			     it != trx->lock.table_locks.end();
-			     it++) {
-				ut_ad(!*it);
-			}
-		}
+    if (!UT_LIST_GET_LEN(lock.trx_locks))
+      for (auto l : lock.table_locks)
+        ut_ad(!l);
 #endif /* UNIV_DEBUG */
-		trx->commit_state();
+    commit_state();
 
-		if (trx->id) {
-			trx_sys.deregister_rw(trx);
+    if (id)
+    {
+      trx_sys.deregister_rw(this);
 
-			/* Wait for any implicit-to-explicit lock
-			conversions to cease, so that there will be no
-			race condition in lock_release(). */
-			while (UNIV_UNLIKELY(trx->is_referenced())) {
-				ut_delay(srv_spin_wait_delay);
-			}
+      /* Wait for any implicit-to-explicit lock conversions to cease,
+      so that there will be no race condition in lock_release(). */
+      while (UNIV_UNLIKELY(is_referenced()))
+        ut_delay(srv_spin_wait_delay);
+      release_locks();
+      id= 0;
+    }
+    else
+    {
+      ut_ad(read_only || !rsegs.m_redo.rseg);
+      release_locks();
+    }
 
-			trx->release_locks();
-			trx->id = 0;
-		} else {
-			ut_ad(trx->read_only || !trx->rsegs.m_redo.rseg);
-			trx->release_locks();
-		}
+    DEBUG_SYNC_C("after_trx_committed_in_memory");
 
-		DEBUG_SYNC_C("after_trx_committed_in_memory");
+    if (read_only || !rsegs.m_redo.rseg)
+    {
+      MONITOR_INC(MONITOR_TRX_RO_COMMIT);
+    }
+    else
+    {
+      trx_update_mod_tables_timestamp(this);
+      MONITOR_INC(MONITOR_TRX_RW_COMMIT);
+      is_recovered= false;
+    }
 
-		if (trx->read_only || !trx->rsegs.m_redo.rseg) {
-			MONITOR_INC(MONITOR_TRX_RO_COMMIT);
-		} else {
-			trx_update_mod_tables_timestamp(trx);
-			MONITOR_INC(MONITOR_TRX_RW_COMMIT);
-			trx->is_recovered = false;
-		}
-	}
+    while (dict_table_t *table= UT_LIST_GET_FIRST(lock.evicted_tables))
+    {
+      UT_LIST_REMOVE(lock.evicted_tables, table);
+      dict_mem_table_free(table);
+    }
+  }
 
-	ut_ad(!trx->rsegs.m_redo.undo);
+  ut_ad(!rsegs.m_redo.undo);
+  ut_ad(UT_LIST_GET_LEN(lock.evicted_tables) == 0);
 
-	if (trx_rseg_t*	rseg = trx->rsegs.m_redo.rseg) {
-		mutex_enter(&rseg->mutex);
-		ut_ad(rseg->trx_ref_count > 0);
-		--rseg->trx_ref_count;
-		mutex_exit(&rseg->mutex);
+  if (trx_rseg_t *rseg= rsegs.m_redo.rseg)
+  {
+    mutex_enter(&rseg->mutex);
+    ut_ad(rseg->trx_ref_count > 0);
+    --rseg->trx_ref_count;
+    mutex_exit(&rseg->mutex);
 
-		if (trx_undo_t*& insert = trx->rsegs.m_redo.old_insert) {
-			ut_ad(insert->rseg == rseg);
-			trx_undo_commit_cleanup(insert, false);
-			insert = NULL;
-		}
-	}
+    if (trx_undo_t *&insert= rsegs.m_redo.old_insert)
+    {
+      ut_ad(insert->rseg == rseg);
+      trx_undo_commit_cleanup(insert, false);
+      insert= nullptr;
+    }
+  }
 
-	ut_ad(!trx->rsegs.m_redo.old_insert);
+  ut_ad(!rsegs.m_redo.old_insert);
 
-	if (mtr != NULL) {
-		if (trx_undo_t*& undo = trx->rsegs.m_noredo.undo) {
-			ut_ad(undo->rseg == trx->rsegs.m_noredo.rseg);
-			trx_undo_commit_cleanup(undo, true);
-			undo = NULL;
-		}
+  if (mtr)
+  {
+    if (trx_undo_t *&undo= rsegs.m_noredo.undo)
+    {
+      ut_ad(undo->rseg == rsegs.m_noredo.rseg);
+      trx_undo_commit_cleanup(undo, true);
+      undo= nullptr;
+    }
 
-		/* NOTE that we could possibly make a group commit more
-		efficient here: call os_thread_yield here to allow also other
-		trxs to come to commit! */
+    /* NOTE that we could possibly make a group commit more efficient
+    here: call os_thread_yield here to allow also other trxs to come
+    to commit! */
 
-		/*-------------------------------------*/
+    /*-------------------------------------*/
 
-		/* Depending on the my.cnf options, we may now write the log
-		buffer to the log files, making the transaction durable if
-		the OS does not crash. We may also flush the log files to
-		disk, making the transaction durable also at an OS crash or a
-		power outage.
+    /* Depending on the my.cnf options, we may now write the log
+    buffer to the log files, making the transaction durable if the OS
+    does not crash. We may also flush the log files to disk, making
+    the transaction durable also at an OS crash or a power outage.
 
-		The idea in InnoDB's group commit is that a group of
-		transactions gather behind a trx doing a physical disk write
-		to log files, and when that physical write has been completed,
-		one of those transactions does a write which commits the whole
-		group. Note that this group commit will only bring benefit if
-		there are > 2 users in the database. Then at least 2 users can
-		gather behind one doing the physical log write to disk.
+    The idea in InnoDB's group commit is that a group of transactions
+    gather behind a trx doing a physical disk write to log files, and
+    when that physical write has been completed, one of those
+    transactions does a write which commits the whole group. Note that
+    this group commit will only bring benefit if there are > 2 users
+    in the database. Then at least 2 users can gather behind one doing
+    the physical log write to disk.
 
-		If we are calling trx_commit() under prepare_commit_mutex, we
-		will delay possible log write and flush to a separate function
-		trx_commit_complete_for_mysql(), which is only called when the
-		thread has released the mutex. This is to make the
-		group commit algorithm to work. Otherwise, the prepare_commit
-		mutex would serialize all commits and prevent a group of
-		transactions from gathering. */
+    If we are calling trx_t::commit() under prepare_commit_mutex, we
+    will delay possible log write and flush to a separate function
+    trx_commit_complete_for_mysql(), which is only called when the
+    thread has released the mutex. This is to make the group commit
+    algorithm to work. Otherwise, the prepare_commit mutex would
+    serialize all commits and prevent a group of transactions from
+    gathering. */
 
-		lsn_t	lsn = mtr->commit_lsn();
+    commit_lsn= mtr->commit_lsn();
+    if (!commit_lsn)
+      /* Nothing to be done. */;
+    else if (flush_log_later)
+      /* Do nothing yet */
+      must_flush_log_later= true;
+    else if (srv_flush_log_at_trx_commit)
+      trx_flush_log_if_needed(commit_lsn, this);
 
-		if (lsn == 0) {
-			/* Nothing to be done. */
-		} else if (trx->flush_log_later) {
-			/* Do nothing yet */
-			trx->must_flush_log_later = true;
-		} else if (srv_flush_log_at_trx_commit == 0) {
-			/* Do nothing */
-		} else {
-			trx_flush_log_if_needed(lsn, trx);
-		}
+    /* Tell server some activity has happened, since the trx does
+    changes something. Background utility threads like master thread,
+    purge thread or page_cleaner thread might have some work to do. */
+    srv_active_wake_master_thread();
+  }
 
-		trx->commit_lsn = lsn;
+  ut_ad(!rsegs.m_noredo.undo);
 
-		/* Tell server some activity has happened, since the trx
-		does changes something. Background utility threads like
-		master thread, purge thread or page_cleaner thread might
-		have some work to do. */
-		srv_active_wake_master_thread();
-	}
+  /* Free all savepoints, starting from the first. */
+  trx_named_savept_t *savep= UT_LIST_GET_FIRST(trx_savepoints);
 
-	ut_ad(!trx->rsegs.m_noredo.undo);
+  trx_roll_savepoints_free(this, savep);
 
-	/* Free all savepoints, starting from the first. */
-	trx_named_savept_t*	savep = UT_LIST_GET_FIRST(trx->trx_savepoints);
+  if (fts_trx)
+    trx_finalize_for_fts(this, undo_no != 0);
 
-	trx_roll_savepoints_free(trx, savep);
+  trx_mutex_enter(this);
+  dict_operation= TRX_DICT_OP_NONE;
+  lock.was_chosen_as_deadlock_victim= false;
 
-        if (trx->fts_trx != NULL) {
-                trx_finalize_for_fts(trx, trx->undo_no != 0);
-        }
+  DBUG_LOG("trx", "Commit in memory: " << this);
+  state= TRX_STATE_NOT_STARTED;
 
-	trx_mutex_enter(trx);
-	trx->dict_operation = TRX_DICT_OP_NONE;
-	trx->lock.was_chosen_as_deadlock_victim = false;
-
-	DBUG_LOG("trx", "Commit in memory: " << trx);
-	trx->state = TRX_STATE_NOT_STARTED;
 #ifdef WITH_WSREP
-	trx->wsrep = false;
-#endif
+  /* Serialization history has been written and the transaction is
+  committed in memory, which makes this commit ordered. Release commit
+  order critical section. */
+  if (wsrep)
+  {
+    wsrep= false;
+    wsrep_commit_ordered(mysql_thd);
+  }
+#endif /* WITH_WSREP */
 
-	assert_trx_is_free(trx);
+  assert_trx_is_free(this);
+  trx_init(this);
+  trx_mutex_exit(this);
 
-	trx_init(trx);
-
-	trx_mutex_exit(trx);
-
-	ut_a(trx->error_state == DB_SUCCESS);
-	srv_wake_purge_thread_if_not_active();
+  ut_a(error_state == DB_SUCCESS);
+  if (!srv_read_only_mode)
+    srv_wake_purge_thread_if_not_active();
 }
 
-/** Commit a transaction and a mini-transaction.
-@param[in,out]	trx	transaction
-@param[in,out]	mtr	mini-transaction (NULL if no modifications) */
-void trx_commit_low(trx_t* trx, mtr_t* mtr)
+/** Commit the transaction in a mini-transaction.
+@param mtr  mini-transaction (if there are any persistent modifications) */
+void trx_t::commit_low(mtr_t *mtr)
 {
-	assert_trx_nonlocking_or_in_list(trx);
-	ut_ad(!trx_state_eq(trx, TRX_STATE_COMMITTED_IN_MEMORY));
-	ut_ad(!mtr || mtr->is_active());
-	ut_d(bool aborted = trx->in_rollback
-	     && trx->error_state == DB_DEADLOCK);
-	ut_ad(!mtr == (aborted || !trx->has_logged_or_recovered()));
-	ut_ad(!mtr || !aborted);
+  assert_trx_nonlocking_or_in_list(this);
+  ut_ad(!trx_state_eq(this, TRX_STATE_COMMITTED_IN_MEMORY));
+  ut_ad(!mtr || mtr->is_active());
+  ut_d(bool aborted = in_rollback && error_state == DB_DEADLOCK);
+  ut_ad(!mtr == (aborted || !has_logged_or_recovered()));
+  ut_ad(!mtr || !aborted);
 
-	/* undo_no is non-zero if we're doing the final commit. */
-	if (trx->fts_trx != NULL && trx->undo_no != 0) {
-		dberr_t	error;
-
-		ut_a(!trx_is_autocommit_non_locking(trx));
-
-		error = fts_commit(trx);
-
-		/* FTS-FIXME: Temporarily tolerate DB_DUPLICATE_KEY
-		instead of dying. This is a possible scenario if there
-		is a crash between insert to DELETED table committing
-		and transaction committing. The fix would be able to
-		return error from this function */
-		if (error != DB_SUCCESS && error != DB_DUPLICATE_KEY) {
-			/* FTS-FIXME: once we can return values from this
-			function, we should do so and signal an error
-			instead of just dying. */
-
-			ut_error;
-		}
-	}
+  /* undo_no is non-zero if we're doing the final commit. */
+  if (fts_trx && undo_no)
+  {
+    ut_a(!trx_is_autocommit_non_locking(this));
+    dberr_t error= fts_commit(this);
+    /* FTS-FIXME: Temporarily tolerate DB_DUPLICATE_KEY instead of
+    dying. This is a possible scenario if there is a crash between
+    insert to DELETED table committing and transaction committing. The
+    fix would be able to return error from this function */
+    ut_a(error == DB_SUCCESS || error == DB_DUPLICATE_KEY);
+  }
 
 #ifndef DBUG_OFF
-	const bool debug_sync = trx->mysql_thd && trx->has_logged_persistent();
+  const bool debug_sync= mysql_thd && has_logged_persistent();
 #endif
 
-	if (mtr != NULL) {
-		trx_write_serialisation_history(trx, mtr);
+  if (mtr)
+  {
+    trx_write_serialisation_history(this, mtr);
 
-		/* The following call commits the mini-transaction, making the
-		whole transaction committed in the file-based world, at this
-		log sequence number. The transaction becomes 'durable' when
-		we write the log to disk, but in the logical sense the commit
-		in the file-based data structures (undo logs etc.) happens
-		here.
+    /* The following call commits the mini-transaction, making the
+    whole transaction committed in the file-based world, at this log
+    sequence number. The transaction becomes 'durable' when we write
+    the log to disk, but in the logical sense the commit in the
+    file-based data structures (undo logs etc.) happens here.
 
-		NOTE that transaction numbers, which are assigned only to
-		transactions with an update undo log, do not necessarily come
-		in exactly the same order as commit lsn's, if the transactions
-		have different rollback segments. To get exactly the same
-		order we should hold the kernel mutex up to this point,
-		adding to the contention of the kernel mutex. However, if
-		a transaction T2 is able to see modifications made by
-		a transaction T1, T2 will always get a bigger transaction
-		number and a bigger commit lsn than T1. */
+    NOTE that transaction numbers, which are assigned only to
+    transactions with an update undo log, do not necessarily come in
+    exactly the same order as commit lsn's, if the transactions have
+    different rollback segments. To get exactly the same order we
+    should hold the kernel mutex up to this point, adding to the
+    contention of the kernel mutex. However, if a transaction T2 is
+    able to see modifications made by a transaction T1, T2 will always
+    get a bigger transaction number and a bigger commit lsn than T1. */
 
-		/*--------------*/
-		mtr_commit(mtr);
-
-		DBUG_EXECUTE_IF("ib_crash_during_trx_commit_in_mem",
-				if (trx->has_logged()) {
-					log_write_up_to(mtr->commit_lsn(),
-							true);
-					DBUG_SUICIDE();
-				});
-		/*--------------*/
-	}
+    mtr->commit();
+  }
 #ifndef DBUG_OFF
-	/* In case of this function is called from a stack executing
-	   THD::release_resources -> ...
-              innobase_connection_close() ->
-                     trx_rollback_for_mysql... -> .
-           mysql's thd does not seem to have
-           thd->debug_sync_control defined any longer. However the stack
-           is possible only with a prepared trx not updating any data.
-        */
-	if (debug_sync) {
-		DEBUG_SYNC_C("before_trx_state_committed_in_memory");
-	}
+  if (debug_sync)
+    DEBUG_SYNC_C("before_trx_state_committed_in_memory");
 #endif
 
-	trx_commit_in_memory(trx, mtr);
+  commit_in_memory(mtr);
 }
 
-/****************************************************************//**
-Commits a transaction. */
-void
-trx_commit(
-/*=======*/
-	trx_t*	trx)	/*!< in/out: transaction */
+
+void trx_t::commit()
 {
-	mtr_t*	mtr;
-	mtr_t	local_mtr;
+  mtr_t *mtr= nullptr;
+  mtr_t local_mtr;
 
-	DBUG_EXECUTE_IF("ib_trx_commit_crash_before_trx_commit_start",
-			DBUG_SUICIDE(););
-
-	if (trx->has_logged_or_recovered()) {
-		mtr = &local_mtr;
-		mtr->start();
-	} else {
-
-		mtr = NULL;
-	}
-
-	trx_commit_low(trx, mtr);
+  if (has_logged_or_recovered())
+  {
+    mtr= &local_mtr;
+    local_mtr.start();
+  }
+  commit_low(mtr);
 }
 
 /****************************************************************//**
@@ -1723,11 +1681,8 @@ trx_commit_step(
 		trx_commit_or_rollback_prepare(trx);
 
 		trx->lock.que_state = TRX_QUE_COMMITTING;
-
-		trx_commit(trx);
-
+		trx->commit();
 		ut_ad(trx->lock.wait_thr == NULL);
-
 		trx->lock.que_state = TRX_QUE_RUNNING;
 
 		thr = NULL;
@@ -1765,9 +1720,7 @@ trx_commit_for_mysql(
 	case TRX_STATE_PREPARED:
 	case TRX_STATE_PREPARED_RECOVERED:
 		trx->op_info = "committing";
-
-		trx_commit(trx);
-
+		trx->commit();
 		MONITOR_DEC(MONITOR_TRX_ACTIVE);
 		trx->op_info = "";
 		return(DB_SUCCESS);
@@ -2251,7 +2204,7 @@ static my_bool trx_get_trx_by_xid_callback(rw_trx_hash_element_t *element,
       transaction needs a valid trx->xid for
       invoking trx_sys_update_wsrep_checkpoint(). */
       if (!wsrep_is_wsrep_xid(trx->xid))
-#endif
+#endif /* WITH_WSREP */
       /* Invalidate the XID, so that subsequent calls will not find it. */
       trx->xid->null();
       arg->trx= trx;
